//! Implementation of the Unreliable, unordered, Datagram transmission model
//!
//! AKA: UDP-like, but the datagram can cross the packet-size (MTU) limit.
//!
//! Only fully received datagrams will be delivered to the user, and
//! half-received ones will be discarded after a timeout

use crate::{
    connection::stream::{
        Chunk, Error, Sequence, SequenceEnd, SequenceStart, StreamData,
    },
    enc::Random,
};

use ::std::collections::{BTreeMap, VecDeque};

#[cfg(test)]
mod tests;

#[derive(Debug, PartialEq, Eq)]
enum Fragment {
    Start((Sequence, Sequence)),
    Middle((Sequence, Sequence)),
    End((Sequence, Sequence)),
    Full((Sequence, Sequence)),
    Delivered((Sequence, Sequence)),
}

impl Fragment {
    // FIXME: sequence start/end?
    fn get_seqs(&self) -> (Sequence, Sequence) {
        match self {
            Fragment::Start((f, t))
            | Fragment::Middle((f, t))
            | Fragment::End((f, t))
            | Fragment::Full((f, t))
            | Fragment::Delivered((f, t)) => (*f, *t),
        }
    }
    fn is_start(&self) -> bool {
        match self {
            Fragment::Start(_) | Fragment::Full(_) | Fragment::Delivered(_) => {
                true
            }
            Fragment::End(_) | Fragment::Middle(_) => false,
        }
    }
    fn is_end(&self) -> bool {
        match self {
            Fragment::End(_) | Fragment::Full(_) | Fragment::Delivered(_) => {
                true
            }
            Fragment::Start(_) | Fragment::Middle(_) => false,
        }
    }
}
/*
impl ::std::cmp::PartialEq<Sequence> for Fragment {
    fn eq(&self, other: &Sequence) -> bool {
        self.get_seq() == *other
    }
}

impl ::std::cmp::PartialOrd<Sequence> for Fragment {
    fn partial_cmp(&self, other: &Sequence) -> Option<::std::cmp::Ordering> {
        Some(self.get_seq().cmp(other))
    }
}

impl ::std::cmp::PartialOrd<Fragment> for Fragment {
    fn partial_cmp(&self, other: &Fragment) -> Option<::std::cmp::Ordering> {
        Some(self.get_seq().cmp(&other.get_seq()))
    }
}
impl Ord for Fragment {
    fn cmp(&self, other: &Fragment) -> ::std::cmp::Ordering {
        self.get_seq().cmp(&other.get_seq())
    }
}
*/

type Timer = u64;

pub(crate) struct Uud {
    pub(crate) window_start: SequenceStart,
    window_end: SequenceEnd,
    pivot: u32,
    data: Vec<u8>,
    track: VecDeque<(Fragment, Timer)>,
}

impl Uud {
    pub(crate) fn new(rand: &Random) -> Self {
        let window_len = 1048576; // 1MB. should be enough for anybody. (lol)
        let window_start = SequenceStart(Sequence::new(rand));
        let window_end = SequenceEnd(window_start.0 + (window_len - 1));
        let mut data = Vec::with_capacity(window_len as usize);
        data.resize(data.capacity(), 0);

        Self {
            window_start,
            window_end,
            pivot: window_len,
            data,
            track: VecDeque::with_capacity(4),
        }
    }
    pub(crate) fn with_window_size(rand: &Random, size: u32) -> Self {
        assert!(
            size < Sequence::max().0.0,
            "Max window size is {}",
            Sequence::max().0.0
        );
        let window_len = size;
        let window_start = SequenceStart(Sequence::new(rand));
        let window_end = SequenceEnd(window_start.0 + (window_len - 1));
        let mut data = Vec::with_capacity(window_len as usize);
        data.resize(data.capacity(), 0);

        Self {
            window_start,
            window_end,
            pivot: window_len,
            data,
            track: VecDeque::with_capacity(4),
        }
    }
    pub(crate) fn window_size(&self) -> u32 {
        self.data.len() as u32
    }
    pub(crate) fn recv(&mut self, chunk: Chunk) -> Result<StreamData, Error> {
        let chunk_to = chunk.sequence + chunk.data.len() as u32;
        if !chunk
            .sequence
            .is_between(self.window_start, self.window_end)
        {
            return Err(Error::OutOfWindow);
        }

        // make sure we consider only the bytes inside the sliding window
        let maxlen = ::std::cmp::min(
            chunk.sequence.remaining_window(self.window_end) as usize,
            chunk.data.len(),
        );
        if maxlen == 0 {
            // empty window or empty chunk, but we don't care
            return Err(Error::OutOfWindow);
        }
        let chunk_flag_end: bool;
        if maxlen != chunk.data.len() {
            // we are not considering the full chunk, so
            // make sure the end flag is not set
            chunk_flag_end = false;

            // FIXME: what happens if we "truncate" this chunk now,
            // then we have more space in the window
            // then we receive the same packet again?
        } else {
            chunk_flag_end = chunk.flag_end;
        }
        // translate Sequences to offsets in self.data
        let data = &chunk.data[..maxlen];
        let chunk_to = chunk.sequence + data.len() as u32;
        let mut last_usable = self.window_end.0;
        let mut ret = StreamData::NotReady;
        let mut copy_data_idx_from = 0;
        // FIXME: and on receiving first fragment after the second, or empty
        // track?
        for (idx, (fragment, _)) in self.track.iter_mut().enumerate().rev() {
            let (from, to) = fragment.get_seqs();
            let to_next = to + 1;
            match to_next.cmp_in_window(self.window_start, chunk.sequence) {
                ::core::cmp::Ordering::Equal => {
                    // `chunk` is immediately after `fragment`
                    if !chunk_to.is_between(
                        SequenceStart(to_next),
                        SequenceEnd(last_usable),
                    ) {
                        return Err(Error::Reconstructing);
                    }
                    match fragment {
                        Fragment::Start((_, f_end)) => {
                            if chunk.flag_start {
                                // we can't start a datagram twice.
                                // ignore the data
                                return Err(Error::WrongFlags);
                            }
                            if chunk_flag_end {
                                *fragment = Fragment::Full((
                                    from,
                                    to + (data.len() as u32),
                                ));
                                ret = StreamData::Ready;
                            } else {
                                *f_end += data.len() as u32;
                            }
                            copy_data_idx_from =
                                chunk.sequence.diff_from(self.window_start.0)
                                    as usize;
                        }
                        Fragment::Middle((_, f_end)) => {
                            if chunk.flag_start {
                                // we can't start a datagram twice.
                                // ignore the data
                                return Err(Error::WrongFlags);
                            }
                            if chunk_flag_end {
                                *fragment = Fragment::End((
                                    from,
                                    to + (data.len() as u32),
                                ));
                            } else {
                                *f_end += data.len() as u32;
                            }
                            copy_data_idx_from =
                                chunk.sequence.diff_from(self.window_start.0)
                                    as usize;
                        }
                        Fragment::End(_)
                        | Fragment::Full(_)
                        | Fragment::Delivered(_) => {
                            if !chunk.flag_start {
                                return Err(Error::WrongFlags);
                            }
                            let toinsert = if chunk_flag_end {
                                ret = StreamData::Ready;
                                Fragment::Full((chunk.sequence, chunk_to))
                            } else {
                                Fragment::Start((chunk.sequence, chunk_to))
                            };
                            self.track.insert(idx + 1, (toinsert, 0));
                            copy_data_idx_from =
                                chunk.sequence.diff_from(self.window_start.0)
                                    as usize;
                        }
                    }
                    break;
                }
                ::core::cmp::Ordering::Less => {
                    // there is a data hole between `chunk` and `fragment`

                    if !chunk_to.is_between(
                        SequenceStart(to_next),
                        SequenceEnd(last_usable),
                    ) {
                        return Err(Error::Reconstructing);
                    }
                    let toinsert = if chunk.flag_start {
                        if chunk_flag_end {
                            ret = StreamData::Ready;
                            Fragment::Full((chunk.sequence, chunk_to))
                        } else {
                            Fragment::Start((chunk.sequence, chunk_to))
                        }
                    } else {
                        if chunk_flag_end {
                            Fragment::End((chunk.sequence, chunk_to))
                        } else {
                            Fragment::Middle((chunk.sequence, chunk_to))
                        }
                    };
                    self.track.insert(idx + 1, (toinsert, 0));
                    copy_data_idx_from =
                        chunk.sequence.diff_from(self.window_start.0) as usize;
                    break;
                }
                ::core::cmp::Ordering::Greater => {
                    // to_next > chunk.sequence
                    // `fragment` is too new, need to look at older ones

                    if from.cmp_in_window(self.window_start, chunk.sequence)
                        != ::core::cmp::Ordering::Greater
                    {
                        // to_next > chunk.sequence >= from
                        // overlapping not yet allowed
                        return Err(Error::Reconstructing);
                    }
                    if idx == 0 {
                        // check if we can add before everything
                        if chunk_to == from {
                            if fragment.is_start() {
                                if chunk_flag_end {
                                    // add, don't merge
                                } else {
                                    //fragment.start, but !chunk.end
                                    return Err(Error::WrongFlags);
                                }
                            } else {
                                if chunk_flag_end {
                                    //chunk.end but !fragment.start
                                    return Err(Error::WrongFlags);
                                } else {
                                    if chunk.flag_start {
                                        if fragment.is_end() {
                                            *fragment = Fragment::Full((
                                                chunk.sequence,
                                                to,
                                            ));
                                            ret = StreamData::Ready;
                                        } else {
                                            *fragment = Fragment::Start((
                                                chunk.sequence,
                                                to,
                                            ));
                                        }
                                    } else {
                                        if fragment.is_end() {
                                            *fragment = Fragment::End((
                                                chunk.sequence,
                                                to,
                                            ));
                                        } else {
                                            *fragment = Fragment::Middle((
                                                chunk.sequence,
                                                to,
                                            ));
                                        }
                                    }
                                }
                            }
                            copy_data_idx_from =
                                chunk.sequence.diff_from(self.window_start.0)
                                    as usize;
                            break;
                        }
                        // chunk before fragment
                        let toinsert = if chunk.flag_start {
                            if chunk_flag_end {
                                ret = StreamData::Ready;
                                Fragment::Full((chunk.sequence, chunk_to))
                            } else {
                                Fragment::Start((chunk.sequence, chunk_to))
                            }
                        } else {
                            if chunk_flag_end {
                                Fragment::End((chunk.sequence, chunk_to))
                            } else {
                                Fragment::Middle((chunk.sequence, chunk_to))
                            }
                        };
                        self.track.insert(0, (toinsert, 0));
                        copy_data_idx_from =
                            chunk.sequence.diff_from(self.window_start.0)
                                as usize;
                        break;
                    }
                    last_usable = from - 1;
                }
            }
        }
        let data_idx_from =
            (copy_data_idx_from + self.pivot as usize) % self.data.len();
        let data_idx_to = (data_idx_from + data.len()) % self.data.len();
        if data_idx_from < data_idx_to {
            self.data[data_idx_from..data_idx_to].copy_from_slice(&data);
        } else {
            let data_pivot = self.data.len() - data_idx_from;
            let (first, second) = data.split_at(data_pivot);
            self.data[data_idx_from..].copy_from_slice(&first);
            self.data[..data_idx_to].copy_from_slice(&data);
        }
        Ok(ret)
    }
}

/// Copy of ROB for reference
#[derive(Debug, Clone)]
pub(crate) struct UnreliableUnorderedDatagram {
    pub(crate) window_start: SequenceStart,
    window_end: SequenceEnd,
    pivot: u32,
    data: Vec<u8>,
    missing: Vec<(Sequence, Sequence)>,
}

impl UnreliableUnorderedDatagram {
    pub(crate) fn new(rand: &Random) -> Self {
        let window_len = 1048576; // 1MB. should be enough for anybody. (lol)
        let window_start = SequenceStart(Sequence::new(rand));
        let window_end = SequenceEnd(window_start.0 + (window_len - 1));
        let mut data = Vec::with_capacity(window_len as usize);
        data.resize(data.capacity(), 0);

        Self {
            window_start,
            window_end,
            pivot: window_len,
            data,
            missing: [(window_start.0, window_end.0)].to_vec(),
        }
    }
    pub(crate) fn with_window_size(rand: &Random, size: u32) -> Self {
        assert!(
            size < Sequence::max().0.0,
            "Max window size is {}",
            Sequence::max().0.0
        );
        let window_len = size; // 1MB. should be enough for anybody. (lol)
        let window_start = SequenceStart(Sequence::new(rand));
        let window_end = SequenceEnd(window_start.0 + (window_len - 1));
        let mut data = Vec::with_capacity(window_len as usize);
        data.resize(data.capacity(), 0);

        Self {
            window_start,
            window_end,
            pivot: window_len,
            data,
            missing: [(window_start.0, window_end.0)].to_vec(),
        }
    }
    pub(crate) fn window_size(&self) -> u32 {
        self.data.len() as u32
    }
    pub(crate) fn get(&mut self) -> Vec<u8> {
        if self.missing.len() == 0 {
            let (first, second) = self.data.split_at(self.pivot as usize);
            let mut ret = Vec::with_capacity(self.data.len());
            ret.extend_from_slice(first);
            ret.extend_from_slice(second);
            self.window_start += ret.len() as u32;
            self.window_end = SequenceEnd(Sequence(
                ::core::num::Wrapping::<u32>(ret.len() as u32),
            ));
            self.data.clear();
            return ret;
        }
        let data_len = self.window_start.offset(self.missing[0].0);
        let last_missing_idx = self.missing.len() - 1;
        let mut last_missing = &mut self.missing[last_missing_idx];
        last_missing.1 += data_len as u32;
        self.window_start += data_len as u32;
        self.window_end += data_len as u32;

        let mut ret = Vec::with_capacity(data_len);
        let (first, second) = self.data[..].split_at(self.pivot as usize);
        let first_len = ::core::cmp::min(data_len, first.len());
        let second_len = data_len - first_len;

        ret.extend_from_slice(&first[..first_len]);
        ret.extend_from_slice(&second[..second_len]);

        self.pivot =
            ((self.pivot as usize + data_len) % self.data.len()) as u32;
        ret
    }
    pub(crate) fn recv(&mut self, chunk: Chunk) -> Result<StreamData, Error> {
        if !chunk
            .sequence
            .is_between(self.window_start, self.window_end)
        {
            return Err(Error::OutOfWindow);
        }
        // make sure we consider only the bytes inside the sliding window
        let maxlen = ::std::cmp::min(
            chunk.sequence.remaining_window(self.window_end) as usize,
            chunk.data.len(),
        );
        if maxlen == 0 {
            // empty window or empty chunk, but we don't care
            return Err(Error::OutOfWindow);
        }
        // translate Sequences to offsets in self.data
        let data = &chunk.data[..maxlen];
        let offset = self.window_start.offset(chunk.sequence);
        let offset_end = offset + chunk.data.len() - 1;

        // Find the chunks we are missing that we can copy,
        // and fix the missing tracker
        let mut copy_ranges = Vec::new();
        let mut to_delete = Vec::new();
        let mut to_add = Vec::new();
        // note: the ranges are (INCLUSIVE, INCLUSIVE)
        for (idx, el) in self.missing.iter_mut().enumerate() {
            let missing_from = self.window_start.offset(el.0);
            if missing_from > offset_end {
                break;
            }
            let missing_to = self.window_start.offset(el.1);
            if missing_to < offset {
                continue;
            }
            if missing_from >= offset && missing_from <= offset_end {
                if missing_to <= offset_end {
                    //  [.....chunk.....]
                    //    [..missing..]
                    to_delete.push(idx);
                    copy_ranges.push((missing_from, missing_to));
                } else {
                    //   [....chunk....]
                    //        [...missing...]
                    copy_ranges.push((missing_from, offset_end));
                    el.0 += ((offset_end - missing_from) + 1) as u32;
                }
            } else if missing_from < offset {
                if missing_to > offset_end {
                    //     [..chunk..]
                    //  [....missing....]
                    to_add.push((
                        el.0 + (((offset_end - missing_from) + 1) as u32),
                        el.1,
                    ));
                    el.1 = el.0 + (((offset - missing_from) - 1) as u32);
                    copy_ranges.push((offset, offset_end));
                } else if offset <= missing_to {
                    //       [....chunk....]
                    //   [...missing...]
                    copy_ranges.push((offset, (missing_to - 0)));
                    el.1 = el.0 + (((offset_end - missing_from) - 1) as u32);
                }
            }
        }
        {
            let mut deleted = 0;
            for idx in to_delete.into_iter() {
                self.missing.remove(idx + deleted);
                deleted = deleted + 1;
            }
        }
        self.missing.append(&mut to_add);
        self.missing
            .sort_by(|(from_a, _), (from_b, _)| from_a.0.0.cmp(&from_b.0.0));

        // copy only the missing data
        let (first, second) = self.data[..].split_at_mut(self.pivot as usize);
        for (from, to) in copy_ranges.into_iter() {
            let to = to + 1;
            if from <= first.len() {
                let first_from = from;
                let first_to = ::core::cmp::min(first.len(), to);
                let data_first_from = from - offset;
                let data_first_to = first_to - offset;
                first[first_from..first_to]
                    .copy_from_slice(&data[data_first_from..data_first_to]);

                let second_to = to - first_to;
                let data_second_to = data_first_to + second_to;
                second[..second_to]
                    .copy_from_slice(&data[data_first_to..data_second_to]);
            } else {
                let second_from = from - first.len();
                let second_to = to - first.len();
                let data_from = from - offset;
                let data_to = to - offset;
                second[second_from..second_to]
                    .copy_from_slice(&data[data_from..data_to]);
            }
        }
        if self.missing.len() == 0
            || self.window_start.offset(self.missing[0].0) == 0
        {
            Ok(StreamData::Ready)
        } else {
            Ok(StreamData::NotReady)
        }
    }
}
